@@ -5,10 +5,11 @@ import os
 import signal
 import sqlite3
 import tempfile
+from copy import copy
 from collections import deque
 from contextlib import contextmanager
-from itertools import (chain, dropwhile, groupby, islice, product, takewhile,
-                       tee)
+from itertools import (chain, groupby, islice, product,
+                       dropwhile, takewhile, zip_longest)
 from pathlib import Path, PurePath
 
 
@@ -46,25 +47,24 @@ class Rows:
         if isinstance(src, tuple):
             conn, tname = src
 
-            self._conn = conn
-            self._elts = None
-            self._genfns = [_build_initgen(conn, tname, [])]
-
-            self._insts = [{
+            self._src = src
+            self._history = [{
                 'cmd': 'fetch',
+                'conn': conn,
                 'tname': tname,
+                'genfn': _build_initgen(conn, tname, [])
+
             }]
 
         else:
-            self._conn = None
-            self._elts = list(src)
+            self._src = list(src)
 
             def gen():
-                yield from self._elts
+                yield from self._src
 
-            self._genfns = [gen]
-            self._insts = [{
-                'cmd': 'elts'
+            self._history = [{
+                'cmd': 'read',
+                'genfn': gen
             }]
 
     def __str__(self):
@@ -74,22 +74,74 @@ class Rows:
         return [r[col] for r in self.iter()]
 
     def zip(self, other):
-        pgen = self._genfns[-1]
+        newself = copy(self)
+        pgen = newself._history[-1]['genfn']
 
-        def gen():
+        def gen1():
+            for r1, r2 in zip(pgen(), _rows2iter(other)):
+                yield {**r1, **r2}
+
+        # if the source is the database, it's safe to update rows directly
+        # because tablemap creates new rows everytime you iterate rows
+        def gen2():
             for r1, r2 in zip(pgen(), _rows2iter(other)):
                 r1.update(r2)
                 yield r1
 
-        self._genfns.append(gen)
-        self._insts.append({
+        # Do not use += operator here, it modifies the object
+        newself._history = newself._history + [{
             'cmd': 'zip',
-            'other': other
-        })
-        return self
+            'genfn': gen2 if self._history[0]['cmd'] == 'fetch' else gen1
+        }]
+
+        return newself
+
+    def zip_longest(self, other):
+        newself = copy(self)
+        pgen = newself._history[-1]['genfn']
+
+        def gen1():
+            r1, rs1 = _spy(pgen())
+            r2, rs2 = _spy(_rows2iter(other))
+            empty1 = {col: '' for col in r1.keys()}
+            empty2 = {col: '' for col in r2.keys()}
+
+            for r1, r2 in zip_longest(rs1, rs2):
+                if r1 and r2:
+                    yield {**r1, **r2}
+                elif r1 and not r2:
+                    yield {**r1, **empty2}
+                elif not r1 and r2:
+                    yield {**empty1, **r2}
+
+        def gen2():
+            r1, rs1 = _spy(pgen())
+            r2, rs2 = _spy(_rows2iter(other))
+            # using the same object over and over is SAFE
+            empty1 = {col: '' for col in r1.keys()}
+            empty2 = {col: '' for col in r2.keys()}
+
+            for r1, r2 in zip_longest(rs1, rs2):
+                if r1 and r2:
+                    r1.update(r2)
+                    yield r1
+                elif r1 and not r2:
+                    r1.update(empty2)
+                    yield r1
+                elif not r1 and r2:
+                    empty1.update(r2)
+                    yield empty1
+
+        # Do not use += operator here, it modifies the object
+        newself._history = newself._history + [{
+            'cmd': 'zip_longest',
+            'genfn': gen2 if self._history[0]['cmd'] == 'fetch' else gen1
+        }]
+        return newself
 
     def chain(self, other):
-        pgen = self._genfns[-1]
+        newself = copy(self)
+        pgen = newself._history[-1]['genfn']
 
         def gen():
             for r in pgen():
@@ -97,27 +149,26 @@ class Rows:
             for r in _rows2iter(other):
                 yield r
 
-        self._genfns.append(gen)
-        self._insts.append({
+        newself._history = newself._history + [{
             'cmd': 'chain',
-            'other': other
-        })
-        return self
+            'genfn': gen
+        }]
+        return newself
 
     # each column may contain asc or desc
     # ex) 'col1 desc', 'col2 asc'
     def order(self, *cols):
-        if len(self._insts) == 1 and self._insts[-1]['cmd'] == 'fetch':
-            tname = self._insts[-1]['tname']
-            # replace the previous generator
-            self._genfns[0] = _build_initgen(self._conn, tname, cols)
-            self._insts.append({
-                'cmd': 'order',
-                'cols': cols
-            })
-            return self
+        newself = copy(self)
+        if len(newself._history) == 1\
+                and newself._history[-1]['cmd'] == 'fetch':
+            phist = newself._history[-1]
+            newself._history = [{
+                **phist,
+                'genfn': _build_initgen(phist['conn'], phist['tname'], cols)
+            }]
+            return newself
 
-        pgen = self._genfns[-1]
+        pgen = newself._history[-1]['genfn']
 
         def gen_from_sql():
             try:
@@ -125,6 +176,7 @@ class Rows:
                 with _connect(tmpdb) as c:
                     _insert(c, 'temp', pgen())
                     yield from _fetch(c, 'temp', cols)
+            # TODO: possible errors must be dealt with.
             finally:
                 # must close the file descriptor to delete it
                 os.close(tmpdbfd)
@@ -145,20 +197,19 @@ class Rows:
                 xs.sort(key=lambda r: r[c], reverse=rev)
             yield from xs
 
-        if self._insts[0]['cmd'] == 'fetch':
-            self._genfns.append(gen_from_sql)
-        else:
-            self._genfns.append(gen_simp)
-
-        self._genfns.append(gen_from_sql)
-        self._insts.append({
+        # Do not use += operator here, it modifies the object
+        newself._history = newself._history + [{
             'cmd': 'order',
-            'cols': cols
-        })
-        return self
+            'cols': cols,
+            'genfn': gen_from_sql if newself._history[0]['cmd'] == 'fetch'
+            else gen_simp
+        }]
+
+        return newself
 
     def apply(self, fn):
-        pgen = self._genfns[-1]
+        newself = copy(self)
+        pgen = newself._history[-1]['genfn']
 
         def gen_single():
             for r in pgen():
@@ -172,51 +223,58 @@ class Rows:
             for rs in pgen():
                 yield from fn(rs)
 
-        if self._insts[-1]['cmd'] == 'group':
-            self._genfns.append(gen_group)
-        elif self._insts[-1]['cmd'] == 'windowed':
-            self._genfns.append(gen_windowed)
+        genfn = None
+        if newself._history[-1]['cmd'] == 'group':
+            genfn = gen_group
+        elif newself._history[-1]['cmd'] == 'windowed':
+            genfn = gen_windowed
         else:
-            self._genfns.append(gen_single)
+            genfn = gen_single
 
-        self._insts.append({
+        # Do not use += operator here, it modifies the object
+        newself._history = newself._history + [{
             'cmd': 'apply',
-            'fn': fn
-        })
-        return self
+            'genfn': genfn
+        }]
+        return newself
 
     def group(self, *cols):
-        pgen = self._genfns[-1]
+        newself = copy(self)
+        pgen = newself._history[-1]['genfn']
 
         def gen():
             yield from groupby(pgen(), _keyfn(cols))
 
-        self._genfns.append(gen)
-        self._insts.append({
+         # Do not use += operator here, it modifies the object
+        newself._history = newself._history + [{
             'cmd': 'group',
-            'cols': cols
-        })
-        return self
+            'cols': cols,
+            'genfn': gen
+        }]
+
+        return newself
 
     def _merge(self, fn, other):
-        pgen = self._genfns[-1]
+        newself = copy(self)
+        pgen = newself._history[-1]['genfn']
 
         def gen():
             yield from _step(fn, pgen(), _rows2iter(other))
 
-        self._genfns.append(gen)
-        self._insts.append({
+         # Do not use += operator here, it modifies the object
+        newself._history = newself._history + [{
             'cmd': '_merge',
-            'fn': fn,
-            'other': other
-        })
-        return self
+            'genfn': gen
+        }]
+
+        return newself
 
     def map(self, fn):
-        if self._insts[-1]['cmd'] == 'group':
+        if self._history[-1]['cmd'] == 'group':
             def _fn2gen(f):
                 def gen(_k, rs):
                     x = f(Rows(rs))
+
                     if isinstance(x, dict):
                         yield x
                     elif isinstance(x, Rows):
@@ -224,9 +282,8 @@ class Rows:
                 return gen
             return self.apply(_fn2gen(fn))
 
-        elif self._insts[-1]['cmd'] == 'windowed':
+        elif self._history[-1]['cmd'] == 'windowed':
             def _fn2gen1(f):
-
                 def gen(rs):
                     x = f(Rows(rs))
                     if isinstance(x, dict):
@@ -235,23 +292,7 @@ class Rows:
                         yield from x.iter()
                 return gen
 
-            def _fn2gen2(f):
-                def gen(rs):
-                    rs = rs[:step] + [r.copy() for r in rs[step:]]
-                    x = f(Rows(rs))
-                    if isinstance(x, dict):
-                        yield x
-                    elif isinstance(x, Rows):
-                        yield from x.iter()
-                return gen
-
-            step = self._insts[-1]['step']
-            n = self._insts[-1]['n']
-
-            if step >= n:
-                # no need to copy
-                return self.apply(_fn2gen1(fn))
-            return self.apply(_fn2gen2(fn))
+            return self.apply(_fn2gen1(fn))
 
         # single row
         def _fn2gen_single(f):
@@ -277,15 +318,20 @@ class Rows:
         return self.apply(gen)
 
     def update(self, **kwargs):
-        def updatefn(r):
+        def updatefn1(r):
+            d = r.copy()
             for k, v in kwargs.items():
-                if callable(v):
-                    r.update({k: v(r)})
-                else:
-                    r.update({k: v})
+                d[k] = v(d) if callable(v) else v
+            yield d
+
+        def updatefn2(r):
+            for k, v in kwargs.items():
+                r[k] = v(r) if callable(v) else v
             yield r
 
-        return self.apply(updatefn)
+        if self._history[0]['cmd'] == 'fetch':
+            return self.apply(updatefn2)
+        return self.apply(updatefn1)
 
     # pretty expensive for what it actually does
     # but this version does not depend on the order of rows.
@@ -300,32 +346,26 @@ class Rows:
 
     def fold(self, **kwargs):
         gcols = []
-        if self._insts[-1]['cmd'] == 'group':
-            gcols = self._insts[-1]['cols']
+        if self._history[-1]['cmd'] == 'group':
+            gcols = self._history[-1]['cols']
 
         def foldfn1(rs):
-            ds = Rows(rs)
+            rs = Rows(rs)
             r = {}
             for k, fn in kwargs.items():
-                if callable(fn):
-                    r[k] = fn(ds)
-                else:
-                    r[k] = fn
+                r[k] = fn(rs) if callable(fn) else fn
             yield r
 
         def foldfn2(keys, rs):
-            ds = Rows(rs)
+            rs = Rows(rs)
             r = dict(zip(gcols, keys))
             for k, fn in kwargs.items():
-                if callable(fn):
-                    r[k] = fn(ds)
-                else:
-                    r[k] = fn
+                r[k] = fn(rs) if callable(fn) else fn
             yield r
 
-        if self._insts[-1]['cmd'] == 'windowed':
+        if self._history[-1]['cmd'] == 'windowed':
             return self.apply(foldfn1)
-        if self._insts[-1]['cmd'] == 'group':
+        if self._history[-1]['cmd'] == 'group':
             return self.apply(foldfn2)
         raise ValueError("fold must be preceded by grouping processes")
 
@@ -388,6 +428,10 @@ class Rows:
                 for r1, r2 in product(rs1, rs2):
                     # TODO: updating on the same object multiple times.
                     # Still, why works?!!,
+                    # yielding and returning a appended list is different.
+                    # later modifications do not affect previous yielded object
+                    # So anyway, these updates in all joining methods
+                    # are perfectly safe.
                     r1.update(r2)
                     yield r1
 
@@ -412,73 +456,70 @@ class Rows:
         return self.by(*columns).apply(distinctfn)
 
     def windowed(self, n, step):
-        pgen = self._genfns[-1]
+        newself = copy(self)
+        pgen = newself._history[-1]['genfn']
 
         def gen():
             yield from _windowed(pgen(), n, step)
 
-        self._genfns.append(gen)
-        self._insts.append({
+         # Do not use += operator here, it modifies the object
+        newself._history = newself._history + [{
             'cmd': 'windowed',
             'n': n,
-            'step': step
-        })
-        return self
+            'step': step,
+            'genfn': gen
+        }]
+        return newself
 
     def islice(self, *args):
-        pgen = self._genfns[-1]
+        newself = copy(self)
+        pgen = self._history[-1]['genfn']
 
         def gen():
             yield from islice(pgen(), *args)
 
-        self._genfns.append(gen)
-        self._insts.append({
+         # Do not use += operator here, it modifies the object
+        newself._history = newself._history + [{
             'cmd': 'islice',
-            'args': args
-        })
-        return self
+            'args': args,
+            'genfn': gen
+        }]
+        return newself
 
     def takewhile(self, pred):
-        pgen = self._genfns[-1]
+        newself = copy(self)
+        pgen = newself._history[-1]['genfn']
 
         def gen():
             yield from takewhile(pred, pgen())
 
-        self._genfns.append(gen)
-        self._insts.append({
+        # Do not use += operator here, it modifies the object
+        newself._history = newself._history + [{
             'cmd': 'takewhile',
-            'pred': pred
-        })
-        return self
+            'genfn': gen
+        }]
+        return newself
 
     def dropwhile(self, pred):
-        pgen = self._genfns[-1]
+        newself = copy(self)
+        pgen = newself._history[-1]['genfn']
 
         def gen():
             yield from dropwhile(pred, pgen())
 
-        self._genfns.append(gen)
-        self._insts.append({
+        # Do not use += operator here, it modifies the object
+        newself._history = newself._history + [{
             'cmd': 'dropwhile',
-            'pred': pred
-        })
-        return self
+            'genfn': gen
+        }]
+        return newself
 
     def iter(self):
-        # TODO: analyze self._insts and check if it's semantically ok.
-        yield from self._genfns[-1]()
+        yield from self._history[-1]['genfn']()
 
     def list(self):
         return list(self.iter())
 
-    # experimental
-    def tee(self, n):
-        return tuple(Rows(x) for x in tee(self.iter(), n))
-
-    # experimental
-    def new(self):
-        return Rows([r.copy() for r in self.iter()])
-
 
 def _insert_statement(name, d):
     """insert into foo values (:a, :b, :c, ...)
@@ -500,10 +541,7 @@ def _create_statement(tname, cols):
 
 
 def _dict_factory(cursor, row):
-    d = {}
-    for idx, col in enumerate(cursor.description):
-        d[col[0]] = row[idx]
-    return d
+    return {col[0]: val for col, val in zip(cursor.description, row)}
 
 
 def _keyfn(cols):
