@@ -3,7 +3,7 @@ import unittest
 from itertools import product
 from pathlib import Path
 
-from tablemap import Conn
+from tablemap import Conn, Rows
 
 
 class TestCore(unittest.TestCase):
@@ -36,18 +36,50 @@ class TestCore(unittest.TestCase):
         if Path(self.dbfile1).is_file():
             os.remove(self.dbfile1)
 
+    def test_wierd(self):
+        conn = Conn(self.dbfile)
+        conn['t1'] = self.t1
+
+        xs = conn['t1']
+        self.assertEqual(xs.by('col1').fold(col2_tot=sum(xs['col2'])).list(),
+                         [{'col1': 'a', 'col2_tot': 26},
+                         {'col1': 'b', 'col2_tot': 26},
+                         {'col1': 'c', 'col2_tot': 26},
+                         {'col1': 'd', 'col2_tot': 26}])
+
+    def test_list_mod(self):
+        conn = Conn(self.dbfile)
+        conn['t1'] = self.t1
+
+        def tempfn(rs):
+            rslist = rs.list()
+            for r in rslist:
+                r['col2'] += 1
+
+            return rs
+
+        conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
+        self.assertEqual(conn['t1_1'].list(), [
+            {'col1': 'a', 'col2': 5},
+            {'col1': 'a', 'col2': 6},
+            {'col1': 'b', 'col2': 2},
+            {'col1': 'c', 'col2': 4},
+            {'col1': 'c', 'col2': 5},
+            {'col1': 'c', 'col2': 8},
+            {'col1': 'd', 'col2': 3}])
+
     def test_do_later(self):
         conn = Conn(self.dbfile)
         conn['t1'] = self.t1
 
-        xs, ys = conn['t1'].tee(2)
-        xs.by('col1')
-        tot = sum(ys['col2'])
-        xs.fold(
+        xs = conn['t1']
+        xs1 = xs.by('col1')
+        tot = sum(xs['col2'])
+        xs1 = xs1.fold(
             col3=lambda rs: sum(rs['col2']),
             col4=tot
         )
-        self.assertEqual(xs.list(),
+        self.assertEqual(xs1.list(),
                          [{'col1': 'a', 'col3': 9, 'col4': 26},
                           {'col1': 'b', 'col3': 1, 'col4': 26},
                           {'col1': 'c', 'col3': 14, 'col4': 26},
@@ -73,32 +105,71 @@ class TestCore(unittest.TestCase):
                           {'col1': 'd', 'col2': 2, 'n': 1}]
                          )
 
+    def test_zip_longest(self):
+        conn = Conn(self.dbfile)
+        conn['t1'] = self.t1
+
+        def tempfn(rs):
+            return rs.zip_longest([
+                {'n': 1},
+                {'n': 2}
+            ])
+
+        conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
+        self.assertEqual(conn['t1_1'].list(),
+                         [{'col1': 'a', 'col2': 4, 'n': 1},
+                          {'col1': 'a', 'col2': 5, 'n': 2},
+                          {'col1': 'b', 'col2': 1, 'n': 1},
+                          {'col1': '', 'col2': '', 'n': 2},
+                          {'col1': 'c', 'col2': 3, 'n': 1},
+                          {'col1': 'c', 'col2': 4, 'n': 2},
+                          {'col1': 'c', 'col2': 7, 'n': ''},
+                          {'col1': 'd', 'col2': 2, 'n': 1},
+                          {'col1': '', 'col2': '', 'n': 2}]
+                         )
+
+    def test_zip_longest2(self):
+        conn = Conn(self.dbfile)
+        conn['t1'] = self.t1
+
+        conn['t1_1'] = conn['t1'].zip_longest(({'n': i} for i in range(10)))
+        self.assertEqual(conn['t1_1'].list(),
+                         [{'col1': 'a', 'col2': 4, 'n': 0},
+                          {'col1': 'a', 'col2': 5, 'n': 1},
+                          {'col1': 'b', 'col2': 1, 'n': 2},
+                          {'col1': 'c', 'col2': 3, 'n': 3},
+                          {'col1': 'c', 'col2': 4, 'n': 4},
+                          {'col1': 'c', 'col2': 7, 'n': 5},
+                          {'col1': 'd', 'col2': 2, 'n': 6},
+                          {'col1': '', 'col2': '', 'n': 7},
+                          {'col1': '', 'col2': '', 'n': 8},
+                          {'col1': '', 'col2': '', 'n': 9}])
+
+        conn['t1_2'] = conn['t1'].zip_longest(({'n': i} for i in range(5)))
+        self.assertEqual(conn['t1_2'].list(),
+                         [{'col1': 'a', 'col2': 4, 'n': 0},
+                          {'col1': 'a', 'col2': 5, 'n': 1},
+                          {'col1': 'b', 'col2': 1, 'n': 2},
+                          {'col1': 'c', 'col2': 3, 'n': 3},
+                          {'col1': 'c', 'col2': 4, 'n': 4},
+                          {'col1': 'c', 'col2': 7, 'n': ''},
+                          {'col1': 'd', 'col2': 2, 'n': ''}])
+
     def test_tee_n_new(self):
         conn = Conn(self.dbfile)
 
         def tempfn(rs):
-            rs1, rs2 = rs.tee(2)
-            return rs1.order('col2 desc').chain(rs2.order('col2'))
+            return rs.order('col2 desc').chain(rs.order('col2'))
+        conn['t10'] = self.t1
+
+        conn['t1_1'] = conn['t10'].by('col1').map(tempfn)
 
-        conn['t1'] = self.t1
-        conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
         self.assertEqual(conn['t1_1']['col2'],
                          [5, 4, 4, 5, 1, 1, 7, 4, 3, 3, 4, 7, 2, 2])
 
         def tempfn(rs):
-            rs1, rs2 = rs.tee(2)
-            rs1.update(col2=lambda r: r['col2'] + 1)
-            return rs1.order('col2 desc').chain(rs2.order('col2'))
-
-        conn['t1'] = self.t1
-        conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
-        self.assertEqual(conn['t1_1']['col2'],
-                         [6, 5, 5, 6, 2, 2, 8, 5, 4, 4, 5, 8, 3, 3])
-
-        def tempfn(rs):
-            rs1, rs2 = rs.tee(2)
-            rs1.update(col2=lambda r: r['col2'] + 1)
-            return rs1.order('col2 desc').chain(rs2.new().order('col2'))
+            rs1 = rs.update(col2=lambda r: r['col2'] + 1)
+            return rs1.order('col2 desc').chain(rs.order('col2'))
 
         conn['t1'] = self.t1
         conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
@@ -109,9 +180,8 @@ class TestCore(unittest.TestCase):
         conn = Conn(self.dbfile)
 
         def tempfn(rs):
-            rs1, rs2 = rs.tee(2)
-            return rs1.order('col2').zip(rs2.order('col2 desc')
-                                         .rename(col2_1='col2'))
+            return rs.order('col2').zip(rs.order('col2 desc')
+                                        .rename(col2_1='col2'))
 
         conn['t1'] = self.t1
         conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
@@ -125,30 +195,12 @@ class TestCore(unittest.TestCase):
         self.assertEqual(conn['t1_1'].list(), xs)
 
         def tempfn1(rs):
-            rs1, rs2 = rs.tee(2)
-            rs1.update(col2=lambda r: r['col2'] + 1)
-            return rs1.order('col2').zip(rs2.order('col2 desc')
+            rs1 = rs.update(col2=lambda r: r['col2'] + 1)
+            return rs1.order('col2').zip(rs.order('col2 desc')
                                          .rename(col2_1='col2'))
 
         conn['t1_1'] = conn['t1'].by('col1').map(tempfn1)
 
-        xs = [{'col1': 'a', 'col2': 5, 'col2_1': 6},
-              {'col1': 'a', 'col2': 6, 'col2_1': 5},
-              {'col1': 'b', 'col2': 2, 'col2_1': 2},
-              {'col1': 'c', 'col2': 4, 'col2_1': 8},
-              {'col1': 'c', 'col2': 5, 'col2_1': 5},
-              {'col1': 'c', 'col2': 8, 'col2_1': 4},
-              {'col1': 'd', 'col2': 3, 'col2_1': 3}]
-        self.assertEqual(conn['t1_1'].list(), xs)
-
-        def tempfn2(rs):
-            rs1, rs2 = rs.tee(2)
-            rs1.update(col2=lambda r: r['col2'] + 1)
-            return rs1.order('col2').zip(rs2.new().order('col2 desc')
-                                         .rename(col2_1='col2'))
-
-        conn['t1_1'] = conn['t1'].by('col1').map(tempfn2)
-
         xs = [{'col1': 'a', 'col2': 5, 'col2_1': 5},
               {'col1': 'a', 'col2': 6, 'col2_1': 4},
               {'col1': 'b', 'col2': 2, 'col2_1': 1},
@@ -156,6 +208,7 @@ class TestCore(unittest.TestCase):
               {'col1': 'c', 'col2': 5, 'col2_1': 4},
               {'col1': 'c', 'col2': 8, 'col2_1': 3},
               {'col1': 'd', 'col2': 3, 'col2_1': 2}]
+
         self.assertEqual(conn['t1_1'].list(), xs)
 
     def test_index(self):
@@ -393,9 +446,6 @@ class TestCore(unittest.TestCase):
         conn = Conn(self.dbfile)
         conn['t1'] = self.t1
 
-        # for r in conn['t1'].iter():
-        #     print(r)
-
         conn['t1_1'] = conn['t1'].update(
             col2=lambda r: r['col2'] + 1,
             col3=lambda r: r['col1'] + str(r['col2'])
@@ -449,6 +499,15 @@ class TestCore(unittest.TestCase):
             'col1': 'd', 'col2': 19}]).by('col1')\
             .full_join(conn['t2'].by('col1'))
 
+        conn['t1_1'] = conn['t1'].chain([{'col1': 'd', 'col2': 9}, {
+            'col1': 'd', 'col2': 19}]).by('col1')\
+            .full_join(conn['t2'].by('col1'))
+
+        conn['t1_2'] = conn['t1'].chain([{'col1': 'd', 'col2': 9}, {
+            'col1': 'd', 'col2': 19}]).by('col1')\
+            .full_join(conn['t2'].by('col1'))\
+            .update(col4=lambda r: r['col3'] + 1 if r['col3'] else '')
+
         self.assertEqual(conn['t1_1'].list(),
                          [{'col1': 'a', 'col2': 4, 'col3': ''},
                           {'col1': 'a', 'col2': 5, 'col3': ''},
